from .datasets.dataset import Dataset
from .datasets.fixtures import ExampleDatasets, load_example
from .datasets.schema import EmbeddingColumnNames, RetrievalEmbeddingColumnNames, Schema
from .session.session import Session, active_session, close_app, launch_app

__version__ = "0.0.29rc7"

# module level doc-string
__doc__ = """
arize-phoenix - ML Observability in a notebook
=====================================================================
**phoenix** is a Python package that provides MLOps insights at
lightning speed with zero-config observability for model drift, performance, and
data quality.

Main Features
-------------
Here are just a few of the things that phoenix does well:
  - Compare two sets of model inferences against one another
  - Identify problematic embeddings cohorts using UMAP and clustering
  - Explore model performance, drift, and data quality metrics
"""

__all__ = [
    "Dataset",
    "EmbeddingColumnNames",
    "RetrievalEmbeddingColumnNames",
    "Schema",
    "load_example",
    "ExampleDatasets",
    "active_session",
    "close_app",
    "launch_app",
    "Session",
]



from enum import Enum
from typing import Dict, Union

import strawberry

from phoenix.core.model_schema import PRIMARY, REFERENCE


@strawberry.enum
class DatasetRole(Enum):
    primary = PRIMARY
    reference = REFERENCE


class AncillaryDatasetRole(Enum):
    corpus = "DatasetRole.CORPUS"


STR_TO_DATASET_ROLE: Dict[str, Union[DatasetRole, AncillaryDatasetRole]] = {
    str(DatasetRole.primary.value): DatasetRole.primary,
    str(DatasetRole.reference.value): DatasetRole.reference,
    str(AncillaryDatasetRole.corpus.value): AncillaryDatasetRole.corpus,
}



import atexit
import errno
import logging
import os
from argparse import ArgumentParser
from pathlib import Path
from typing import Optional

import uvicorn

import phoenix.config as config
from phoenix.core.model_schema_adapter import create_model_from_datasets
from phoenix.datasets.dataset import Dataset
from phoenix.datasets.fixtures import FIXTURES, get_datasets
from phoenix.server.app import create_app

logger = logging.getLogger(__name__)


def _write_pid_file() -> None:
    with open(_get_pid_file(), "w"):
        pass


def _remove_pid_file() -> None:
    try:
        os.unlink(_get_pid_file())
    except OSError as e:
        if e.errno == errno.ENOENT:
            # If the pid file doesn't exist, ignore and continue on since
            # we are already in the desired end state; This should not happen
            pass
        else:
            raise


def _get_pid_file() -> str:
    return os.path.join(config.get_pids_path(), "%d" % os.getpid())


if __name__ == "__main__":
    primary_dataset_name: str
    reference_dataset_name: Optional[str]
    # automatically remove the pid file when the process is being gracefully terminated
    atexit.register(_remove_pid_file)
    _write_pid_file()

    parser = ArgumentParser()
    parser.add_argument("--export_path")
    parser.add_argument("--port", type=int, default=config.PORT)
    parser.add_argument("--no-internet", action="store_true")
    parser.add_argument("--debug", action="store_false")  # TODO: Disable before public launch
    subparsers = parser.add_subparsers(dest="command", required=True)
    datasets_parser = subparsers.add_parser("datasets")
    datasets_parser.add_argument("--primary", type=str, required=True)
    datasets_parser.add_argument("--reference", type=str, required=False)
    datasets_parser.add_argument("--corpus", type=str, required=False)
    fixture_parser = subparsers.add_parser("fixture")
    fixture_parser.add_argument("fixture", type=str, choices=[fixture.name for fixture in FIXTURES])
    fixture_parser.add_argument("--primary-only", type=bool)
    args = parser.parse_args()
    export_path = Path(args.export_path) if args.export_path else config.EXPORT_DIR
    if args.command == "datasets":
        primary_dataset_name = args.primary
        reference_dataset_name = args.reference
        corpus_dataset_name = args.corpus
        primary_dataset = Dataset.from_name(primary_dataset_name)
        reference_dataset = (
            Dataset.from_name(reference_dataset_name)
            if reference_dataset_name is not None
            else None
        )
        corpus_dataset = (
            None if corpus_dataset_name is None else Dataset.from_name(corpus_dataset_name)
        )
    else:
        fixture_name = args.fixture
        primary_only = args.primary_only
        primary_dataset, reference_dataset, corpus_dataset = get_datasets(
            fixture_name,
            args.no_internet,
        )
        if primary_only:
            reference_dataset_name = None
            reference_dataset = None

    model = create_model_from_datasets(
        primary_dataset,
        reference_dataset,
    )
    app = create_app(
        export_path=export_path,
        model=model,
        corpus=None if corpus_dataset is None else create_model_from_datasets(corpus_dataset),
        debug=args.debug,
    )

    uvicorn.run(app, port=args.port)



import math
from typing import List

import numpy as np
from pandas import DataFrame, Series
from pandas.api.types import is_numeric_dtype, is_string_dtype

from . import errors as err
from .schema import EmbeddingColumnNames, Schema

RESERVED_EMBEDDING_NAMES = ("prompt", "response")


def _check_valid_schema(schema: Schema) -> List[err.ValidationError]:
    errs: List[str] = []
    if schema.excluded_column_names is None:
        return []

    if schema.timestamp_column_name in schema.excluded_column_names:
        errs.append(
            f"{schema.timestamp_column_name} cannot be excluded because "
            f"it is already being used as the timestamp column"
        )

    if schema.prediction_id_column_name in schema.excluded_column_names:
        errs.append(
            f"{schema.prediction_id_column_name} cannot be excluded because "
            f"it is already being used as the prediction id column"
        )

    if len(errs) > 0:
        return [err.InvalidSchemaError(errs)]

    return []


def validate_dataset_inputs(dataframe: DataFrame, schema: Schema) -> List[err.ValidationError]:
    errors = _check_missing_columns(dataframe, schema)
    if errors:
        return errors
    errors = _check_column_types(dataframe, schema)
    if errors:
        return errors
    errors = _check_valid_schema(schema)
    if errors:
        return errors
    errors = _check_valid_embedding_data(dataframe, schema)
    if errors:
        return errors
    errors = _check_valid_prompt_response_data(dataframe, schema)
    if errors:
        return errors
    return []


def _check_valid_embedding_data(dataframe: DataFrame, schema: Schema) -> List[err.ValidationError]:
    embedding_col_names = schema.embedding_feature_column_names
    if embedding_col_names is None:
        return []

    embedding_errors: List[err.ValidationError] = []
    for embedding_name, column_names in embedding_col_names.items():
        if embedding_name in RESERVED_EMBEDDING_NAMES:
            embedding_errors += _validate_reserved_embedding_name(embedding_name, schema)
        embedding_errors += _validate_embedding_vector(
            dataframe, embedding_name, column_names.vector_column_name
        )

    return embedding_errors


def _check_valid_prompt_response_data(
    dataframe: DataFrame, schema: Schema
) -> List[err.ValidationError]:
    prompt_response_errors: List[err.ValidationError] = []

    prompt_response_column_names = {
        "prompt": schema.prompt_column_names,
        "response": schema.response_column_names,
    }
    for name, column_names in prompt_response_column_names.items():
        if isinstance(column_names, EmbeddingColumnNames):
            prompt_response_errors += _validate_embedding_vector(
                dataframe, name, column_names.vector_column_name
            )

    return prompt_response_errors


def _validate_reserved_embedding_name(
    embedding_name: str, schema: Schema
) -> List[err.ValidationError]:
    if embedding_name == "prompt" and schema.prompt_column_names is not None:
        return [err.InvalidEmbeddingReservedName(embedding_name, "schema.prompt_column_names")]
    elif embedding_name == "response" and schema.response_column_names is not None:
        return [err.InvalidEmbeddingReservedName(embedding_name, "schema.response_column_names")]
    return []


def _validate_embedding_vector(
    dataframe: DataFrame, name: str, vector_column_name: str
) -> List[err.ValidationError]:
    vector_column = dataframe[vector_column_name]
    errors: List[err.ValidationError] = []
    vector_length = None

    for vector in vector_column:
        vector_is_missing = vector is None or (isinstance(vector, float) and math.isnan(vector))
        if vector_is_missing:
            continue

        # Fail if vector is not of supported iterable type
        if not isinstance(vector, (list, np.ndarray, Series)):
            errors.append(
                err.InvalidEmbeddingVectorDataType(
                    embedding_feature_name=name,
                    vector_column_type=str(type(vector)),
                )
            )
            break

        # Fail if not all elements in every vector are int/floats
        allowed_types = (int, float, np.int16, np.int32, np.float16, np.float32)
        if not all(isinstance(val, allowed_types) for val in vector):
            errors.append(
                err.InvalidEmbeddingVectorValuesDataType(
                    embedding_feature_name=name,
                    vector_column_name=vector_column_name,
                    vector=vector,
                )
            )
            break

        if vector_length is None:
            vector_length = len(vector)

        # Fail if vectors in the dataframe are not of the same length, or of length <= 1
        if len(vector) != vector_length:
            errors.append(
                err.EmbeddingVectorSizeMismatch(
                    name,
                    vector_column_name,
                    [vector_length, len(vector)],
                )
            )
            break
        elif vector_length <= 1:
            errors.append(
                err.InvalidEmbeddingVectorSize(
                    name,
                    vector_column_name,
                    vector_length,
                )
            )
            break
    return errors


def _check_column_types(dataframe: DataFrame, schema: Schema) -> List[err.ValidationError]:
    wrong_type_cols: List[str] = []
    if schema.prediction_id_column_name is not None:
        if not (
            is_numeric_dtype(dataframe.dtypes[schema.prediction_id_column_name])
            or is_string_dtype(dataframe.dtypes[schema.prediction_id_column_name])
        ):
            wrong_type_cols.append(
                f"{schema.prediction_id_column_name} should be a string or numeric type"
            )

    if len(wrong_type_cols) > 0:
        return [err.InvalidColumnType(wrong_type_cols)]
    return []


def _check_missing_columns(dataframe: DataFrame, schema: Schema) -> List[err.ValidationError]:
    # converting to a set first makes the checks run a lot faster
    existing_columns = set(dataframe.columns)
    missing_columns = []

    for field in schema.__dict__:
        if field.endswith("column_name"):
            col = getattr(schema, field)
            if col is not None and col not in existing_columns:
                missing_columns.append(col)

    if schema.feature_column_names is not None:
        for col in schema.feature_column_names:
            if col not in existing_columns:
                missing_columns.append(col)

    if schema.embedding_feature_column_names is not None:
        for emb_col_names in schema.embedding_feature_column_names.values():
            if emb_col_names.vector_column_name not in existing_columns:
                missing_columns.append(emb_col_names.vector_column_name)
            if (
                emb_col_names.raw_data_column_name is not None
                and emb_col_names.raw_data_column_name not in existing_columns
            ):
                missing_columns.append(emb_col_names.raw_data_column_name)
            if (
                emb_col_names.link_to_data_column_name is not None
                and emb_col_names.link_to_data_column_name not in existing_columns
            ):
                missing_columns.append(emb_col_names.link_to_data_column_name)

    for column_names in (schema.prompt_column_names, schema.response_column_names):
        if isinstance(column_names, EmbeddingColumnNames):
            if column_names.vector_column_name not in existing_columns:
                missing_columns.append(column_names.vector_column_name)
            if (
                column_names.raw_data_column_name is not None
                and column_names.raw_data_column_name not in existing_columns
            ):
                missing_columns.append(column_names.raw_data_column_name)
            if (
                column_names.link_to_data_column_name is not None
                and column_names.link_to_data_column_name not in existing_columns
            ):
                missing_columns.append(column_names.link_to_data_column_name)

    if missing_columns:
        return [err.MissingColumns(missing_columns)]

    return []



from collections import defaultdict
from itertools import chain
from typing import Dict, List, Optional, Set, Union

import numpy as np
import numpy.typing as npt
import strawberry
from strawberry import ID, UNSET
from strawberry.types import Info
from typing_extensions import Annotated

from phoenix.pointcloud.clustering import Hdbscan
from phoenix.server.api.helpers import ensure_list
from phoenix.server.api.input_types.ClusterInput import ClusterInput
from phoenix.server.api.types.Cluster import Cluster, to_gql_clusters

from .context import Context
from .input_types import Coordinates
from .types.DatasetRole import AncillaryDatasetRole, DatasetRole
from .types.Dimension import to_gql_dimension
from .types.EmbeddingDimension import (
    DEFAULT_CLUSTER_SELECTION_EPSILON,
    DEFAULT_MIN_CLUSTER_SIZE,
    DEFAULT_MIN_SAMPLES,
    to_gql_embedding_dimension,
)
from .types.Event import create_event_id, unpack_event_id
from .types.ExportEventsMutation import ExportEventsMutation
from .types.Model import Model
from .types.node import GlobalID, Node, from_global_id


@strawberry.type
class Query:
    @strawberry.field
    def model(self) -> Model:
        return Model()

    @strawberry.field
    def node(self, id: GlobalID, info: Info[Context, None]) -> Node:
        type_name, node_id = from_global_id(str(id))
        if type_name == "Dimension":
            dimension = info.context.model.scalar_dimensions[node_id]
            return to_gql_dimension(node_id, dimension)
        elif type_name == "EmbeddingDimension":
            embedding_dimension = info.context.model.embedding_dimensions[node_id]
            return to_gql_embedding_dimension(node_id, embedding_dimension)

        raise Exception(f"Unknown node type: {type}")

    @strawberry.field
    def clusters(
        self,
        clusters: List[ClusterInput],
    ) -> List[Cluster]:
        clustered_events: Dict[str, Set[ID]] = defaultdict(set)
        for i, cluster in enumerate(clusters):
            clustered_events[cluster.id or str(i)].update(cluster.event_ids)
        return to_gql_clusters(
            clustered_events=clustered_events,
        )

    @strawberry.field
    def hdbscan_clustering(
        self,
        info: Info[Context, None],
        event_ids: Annotated[
            List[ID],
            strawberry.argument(
                description="Event ID of the coordinates",
            ),
        ],
        coordinates_2d: Annotated[
            Optional[List[Coordinates.InputCoordinate2D]],
            strawberry.argument(
                description="Point coordinates. Must be either 2D or 3D.",
            ),
        ] = UNSET,
        coordinates_3d: Annotated[
            Optional[List[Coordinates.InputCoordinate3D]],
            strawberry.argument(
                description="Point coordinates. Must be either 2D or 3D.",
            ),
        ] = UNSET,
        min_cluster_size: Annotated[
            int,
            strawberry.argument(
                description="HDBSCAN minimum cluster size",
            ),
        ] = DEFAULT_MIN_CLUSTER_SIZE,
        cluster_min_samples: Annotated[
            int,
            strawberry.argument(
                description="HDBSCAN minimum samples",
            ),
        ] = DEFAULT_MIN_SAMPLES,
        cluster_selection_epsilon: Annotated[
            float,
            strawberry.argument(
                description="HDBSCAN cluster selection epsilon",
            ),
        ] = DEFAULT_CLUSTER_SELECTION_EPSILON,
    ) -> List[Cluster]:
        coordinates_3d = ensure_list(coordinates_3d)
        coordinates_2d = ensure_list(coordinates_2d)

        if len(coordinates_3d) > 0 and len(coordinates_2d) > 0:
            raise ValueError("must specify only one of 2D or 3D coordinates")

        if len(coordinates_3d) > 0:
            coordinates = list(
                map(
                    lambda coord: np.array(
                        [coord.x, coord.y, coord.z],
                    ),
                    coordinates_3d,
                )
            )
        else:
            coordinates = list(
                map(
                    lambda coord: np.array(
                        [coord.x, coord.y],
                    ),
                    coordinates_2d,
                )
            )

        if len(event_ids) != len(coordinates):
            raise ValueError(
                f"length mismatch between "
                f"event_ids ({len(event_ids)}) "
                f"and coordinates ({len(coordinates)})"
            )

        if len(event_ids) == 0:
            return []

        grouped_event_ids: Dict[
            Union[DatasetRole, AncillaryDatasetRole],
            List[ID],
        ] = defaultdict(list)
        grouped_coordinates: Dict[
            Union[DatasetRole, AncillaryDatasetRole],
            List[npt.NDArray[np.float64]],
        ] = defaultdict(list)

        for event_id, coordinate in zip(event_ids, coordinates):
            row_id, dataset_role = unpack_event_id(event_id)
            grouped_coordinates[dataset_role].append(coordinate)
            grouped_event_ids[dataset_role].append(create_event_id(row_id, dataset_role))

        stacked_event_ids = (
            grouped_event_ids[DatasetRole.primary]
            + grouped_event_ids[DatasetRole.reference]
            + grouped_event_ids[AncillaryDatasetRole.corpus]
        )
        stacked_coordinates = np.stack(
            chain(
                grouped_coordinates[DatasetRole.primary],
                grouped_coordinates[DatasetRole.reference],
                grouped_coordinates[AncillaryDatasetRole.corpus],
            )
        )  # type: ignore

        clusters = Hdbscan(
            min_cluster_size=min_cluster_size,
            min_samples=cluster_min_samples,
            cluster_selection_epsilon=cluster_selection_epsilon,
        ).find_clusters(stacked_coordinates)

        clustered_events = {
            str(i): {stacked_event_ids[row_idx] for row_idx in cluster}
            for i, cluster in enumerate(clusters)
        }

        return to_gql_clusters(
            clustered_events=clustered_events,
        )


@strawberry.type
class Mutation(ExportEventsMutation):
    ...


schema = strawberry.Schema(query=Query, mutation=Mutation)



from enum import Enum
from functools import partial

import strawberry

from phoenix.metrics.metrics import Cardinality, Count, Max, Mean, Min, PercentEmpty, Quantile, Sum


@strawberry.enum
class DataQualityMetric(Enum):
    cardinality = Cardinality
    percentEmpty = PercentEmpty
    mean = Mean
    sum = Sum
    min = Min
    max = Max
    count = Count
    p01 = partial(Quantile, probability=0.01)
    p25 = partial(Quantile, probability=0.25)
    p50 = partial(Quantile, probability=0.50)
    p75 = partial(Quantile, probability=0.75)
    p99 = partial(Quantile, probability=0.99)



import math
from dataclasses import dataclass
from typing import Any, List, Optional, Union, overload

import numpy as np
import pandas as pd
import strawberry
from strawberry import UNSET

from .DatasetValues import DatasetValues
from .NumericRange import NumericRange


@strawberry.type
class NominalBin:
    """A bin that contains a discrete value"""

    name: str


@strawberry.type
class MissingValueBin:
    name: Optional[str] = UNSET


@strawberry.type
class IntervalBin:
    """A bin that contains a discrete value"""

    # TODO figure out the empty case
    range: NumericRange


@dataclass(frozen=True)
class GqlBinFactory:
    numeric_lbound: float = np.NINF
    numeric_ubound: float = np.inf

    @overload
    def __call__(self, bin: "pd.Interval[float]") -> IntervalBin:
        ...

    @overload
    def __call__(self, bin: Union[str, int, float]) -> Union[NominalBin, MissingValueBin]:
        ...

    def __call__(self, bin: Any) -> Union[NominalBin, IntervalBin, MissingValueBin]:
        if isinstance(bin, pd.Interval):
            return IntervalBin(
                range=NumericRange(
                    start=max(bin.left, self.numeric_lbound),
                    end=min(bin.right, self.numeric_ubound),
                )
            )
        if isinstance(bin, float) and math.isnan(bin):
            return MissingValueBin()
        return NominalBin(name=str(bin))


@strawberry.type
class Segment:
    """A segment of the parent's data, split out using a heuristic"""

    bin: strawberry.union(  # type: ignore
        "Bin",
        types=(
            NominalBin,
            IntervalBin,
            MissingValueBin,
        ),
    )
    counts: DatasetValues = strawberry.field(
        default_factory=DatasetValues,
    )
    # TODO add support for a "z" metric list
    # values: List[Optional[float]]


@strawberry.type
class Segments:
    segments: List[Segment] = strawberry.field(default_factory=list)
    total_counts: DatasetValues = strawberry.field(
        default_factory=DatasetValues,
    )

    def append(self, other: Segment) -> None:
        if (
            isinstance(other.bin, IntervalBin)
            and not math.isfinite(other.bin.range.start)
            and not math.isfinite(other.bin.range.end)
            and other.counts.primary_value == 0
            and other.counts.reference_value == 0
        ):
            # Skip the interval bin with zero counts if it has -inf and +inf
            # as the endpoints, as it could occur due to all values being
            # missing. Because such bin can exist a priori, the caller may
            # still try to append it, but the bin has no real value at this
            # point (and can cause problems for graphql because of the end
            # points are not serializable to JSON).
            return
        self.segments.append(other)
        self.total_counts += other.counts



from collections import defaultdict
from typing import Any, Dict, List, Optional

import pandas as pd
import strawberry
from strawberry import UNSET
from strawberry.types import Info
from typing_extensions import Annotated

import phoenix.core.model_schema as ms
from phoenix.core.model_schema import CONTINUOUS, PRIMARY, REFERENCE, ScalarDimension
from phoenix.metrics import binning
from phoenix.metrics.metrics import Count
from phoenix.metrics.timeseries import row_interval_from_sorted_time_index

from ..context import Context
from ..input_types.Granularity import Granularity
from ..input_types.TimeRange import TimeRange
from .DataQualityMetric import DataQualityMetric
from .DatasetRole import DatasetRole
from .DatasetValues import DatasetValues
from .DimensionDataType import DimensionDataType
from .DimensionShape import DimensionShape
from .DimensionType import DimensionType
from .node import Node
from .ScalarDriftMetricEnum import ScalarDriftMetric
from .Segments import (
    GqlBinFactory,
    Segment,
    Segments,
)
from .TimeSeries import (
    DataQualityTimeSeries,
    DriftTimeSeries,
    ensure_timeseries_parameters,
    get_data_quality_timeseries_data,
    get_drift_timeseries_data,
)


@strawberry.type
class Dimension(Node):
    name: str = strawberry.field(description="The name of the dimension (a.k.a. the column name)")
    type: DimensionType = strawberry.field(
        description="Whether the dimension represents a feature, tag, prediction, or actual."
    )
    dataType: DimensionDataType = strawberry.field(
        description="The data type of the column. Categorical or numeric."
    )
    shape: DimensionShape = strawberry.field(
        description="Whether the dimension data is continuous or discrete."
    )
    dimension: strawberry.Private[ScalarDimension]

    @strawberry.field
    def drift_metric(
        self,
        info: Info[Context, None],
        metric: ScalarDriftMetric,
        time_range: Optional[TimeRange] = UNSET,
    ) -> Optional[float]:
        """
        Computes a drift metric between all reference data and the primary data
        belonging to the input time range (inclusive of the time range start and
        exclusive of the time range end). Returns None if no reference dataset
        exists, if no primary data exists in the input time range, or if the
        input time range is invalid.
        """
        model = info.context.model
        if model[REFERENCE].empty:
            return None
        dataset = model[PRIMARY]
        time_range, granularity = ensure_timeseries_parameters(
            dataset,
            time_range,
        )
        data = get_drift_timeseries_data(
            self.dimension,
            metric,
            time_range,
            granularity,
        )
        return data[0].value if len(data) else None

    @strawberry.field
    async def data_quality_metric(
        self,
        info: Info[Context, None],
        metric: DataQualityMetric,
        time_range: Optional[TimeRange] = UNSET,
        dataset_role: Annotated[
            Optional[DatasetRole],
            strawberry.argument(
                description="The dataset (primary or reference) to query",
            ),
        ] = DatasetRole.primary,
    ) -> Optional[float]:
        if not isinstance(dataset_role, DatasetRole):
            dataset_role = DatasetRole.primary
        dataset = info.context.model[dataset_role.value]
        time_range, granularity = ensure_timeseries_parameters(
            dataset,
            time_range,
        )
        data = get_data_quality_timeseries_data(
            self.dimension,
            metric,
            time_range,
            granularity,
            dataset_role,
        )
        return data[0].value if len(data) else None

    @strawberry.field(
        description=(
            "Returns the observed categories of a categorical dimension (usually a dimension of"
            " string values) as a list of unique string labels sorted in lexicographical order."
            " Missing values are excluded. Non-categorical dimensions return an empty list."
        )
    )  # type: ignore  # https://github.com/strawberry-graphql/strawberry/issues/1929
    def categories(self) -> List[str]:
        return list(self.dimension.categories)

    @strawberry.field(
        description=(
            "Returns the time series of the specified metric for data within a time range. Data"
            " points are generated starting at the end time and are separated by the sampling"
            " interval. Each data point is labeled by the end instant and contains data from their"
            " respective evaluation windows."
        )
    )  # type: ignore  # https://github.com/strawberry-graphql/strawberry/issues/1929
    def data_quality_time_series(
        self,
        info: Info[Context, None],
        metric: DataQualityMetric,
        time_range: TimeRange,
        granularity: Granularity,
        dataset_role: Annotated[
            Optional[DatasetRole],
            strawberry.argument(
                description="The dataset (primary or reference) to query",
            ),
        ] = DatasetRole.primary,
    ) -> DataQualityTimeSeries:
        if not isinstance(dataset_role, DatasetRole):
            dataset_role = DatasetRole.primary
        dataset = info.context.model[dataset_role.value]
        time_range, granularity = ensure_timeseries_parameters(
            dataset,
            time_range,
            granularity,
        )
        return DataQualityTimeSeries(
            data=get_data_quality_timeseries_data(
                self.dimension,
                metric,
                time_range,
                granularity,
                dataset_role,
            )
        )

    @strawberry.field(
        description=(
            "The time series of the specified metric for data within a time range. Data"
            " points are generated starting at the end time and are separated by the sampling"
            " interval. Each data point is labeled by the end instant and contains data from their"
            " respective evaluation windows."
        )
    )  # type: ignore  # https://github.com/strawberry-graphql/strawberry/issues/1929
    def drift_time_series(
        self,
        info: Info[Context, None],
        metric: ScalarDriftMetric,
        time_range: TimeRange,
        granularity: Granularity,
    ) -> DriftTimeSeries:
        model = info.context.model
        if model[REFERENCE].empty:
            return DriftTimeSeries(data=[])
        dataset = model[PRIMARY]
        time_range, granularity = ensure_timeseries_parameters(
            dataset,
            time_range,
            granularity,
        )
        return DriftTimeSeries(
            data=get_drift_timeseries_data(
                self.dimension,
                metric,
                time_range,
                granularity,
            )
        )

    @strawberry.field(
        description="Returns the segments across both datasets and returns the counts per segment",
    )  # type: ignore
    def segments_comparison(
        self,
        info: Info[Context, None],
        primary_time_range: Optional[TimeRange] = UNSET,
    ) -> Segments:
        # TODO: Implement binning across primary and reference

        model = info.context.model
        count = Count()
        summaries = defaultdict(pd.DataFrame)
        binning_method = (
            binning.QuantileBinning(
                reference_series=self.dimension[REFERENCE],
            )
            if self.dimension.data_type is CONTINUOUS
            else binning.CategoricalBinning()
        )
        for role, time_range in (
            (PRIMARY, primary_time_range),
            (REFERENCE, None),
        ):
            if (df := model[role]).empty:
                continue
            if time_range:
                start, stop = row_interval_from_sorted_time_index(
                    df.index,
                    time_range.start,
                    time_range.end,
                )
                df = df.iloc[start:stop]
            summaries[role] = binning_method.segmented_summary(
                self.dimension,
                df,
                (count,),
            )
        segments = Segments()
        lbound, ubound = self.dimension.min_max
        gql_bin_factory = GqlBinFactory(
            numeric_lbound=lbound,
            numeric_ubound=ubound,
        )
        all_bins = summaries[PRIMARY].index.union(summaries[REFERENCE].index)
        if isinstance(binning_method, binning.IntervalBinning) and binning_method.bins is not None:
            all_bins = all_bins.union(binning_method.bins)
        for bin in all_bins:
            values: Dict[ms.DatasetRole, Any] = defaultdict(lambda: None)
            for role in ms.DatasetRole:
                if model[role].empty:
                    continue
                try:
                    result = summaries[role].loc[bin]
                except KeyError:
                    result = {}
                values[role] = count.get_value(result)
            segments.append(
                Segment(
                    bin=gql_bin_factory(bin),
                    counts=DatasetValues(
                        primary_value=values[PRIMARY],
                        reference_value=values[REFERENCE],
                    ),
                )
            )
        return segments


def to_gql_dimension(id_attr: int, dimension: ScalarDimension) -> Dimension:
    """
    Converts a phoenix.core.Dimension to a phoenix.server.api.types.Dimension
    """
    return Dimension(
        id_attr=id_attr,
        name=dimension.name,
        type=DimensionType.from_dimension(dimension),
        dataType=DimensionDataType.from_dimension(dimension),
        dimension=dimension,
        shape=DimensionShape.from_dimension(dimension),
    )



"""
Mixins are behavioral building blocks of metrics. All metrics inherit from
BaseMetric. Other mixins provide specialized functionalities. Mixins rely
on cooperative multiple inheritance and method resolution order in Python.
"""
import collections
import inspect
from abc import ABC, abstractmethod
from dataclasses import dataclass, field, fields, replace
from functools import cached_property
from itertools import repeat
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterator, List, Mapping, Optional

import numpy as np
import pandas as pd
from typing_extensions import TypeAlias

from phoenix.core.model_schema import Column
from phoenix.metrics import Metric
from phoenix.metrics.binning import (
    AdditiveSmoothing,
    BinningMethod,
    CategoricalBinning,
    Normalizer,
)


@dataclass(frozen=True)
class ZeroInitialValue(ABC):
    @property
    def initial_value(self) -> Any:
        if isinstance(self, VectorOperator):
            return np.zeros(self.shape)
        return 0


@dataclass(frozen=True)
class VectorOperator(ABC):
    shape: int = 0


@dataclass(frozen=True)
class NullaryOperator(Metric, ABC):
    def operands(self) -> List[Column]:
        return []


@dataclass(frozen=True)
class UnaryOperator(Metric, ABC):
    """
    A unary operator is a function with one operand or argument as input.
    See https://en.wikipedia.org/wiki/Arity
    """

    operand: Column = Column()

    def operands(self) -> List[Column]:
        return [self.operand]


Actual: TypeAlias = "pd.Series[Any]"
Predicted: TypeAlias = "pd.Series[Any]"

# workaround for type checker
# https://github.com/python/mypy/issues/5446#issuecomment-412043677
if TYPE_CHECKING:
    _BaseMapping = Mapping[str, Any]
else:
    _BaseMapping = collections.abc.Mapping


@dataclass(frozen=True)
class EvaluationMetricKeywordParameters(_BaseMapping):
    """Parameters not compatible with the evaluation function are ignored. For
    example, specifying the positive label for a regression evaluation has no
    effect."""

    pos_label: Optional[Any] = None
    """For classification models, specifying a positive label turns the
    predictions binary. That is, predictions are reassigned a value of True
    when their labels match the positive label, and False otherwise. Missing
    values remain missing and are assigned neither True or False."""
    sample_weight: Optional[Column] = None
    """When applicable for the eval function, a sample weight column is used to
    add weight each observation. E.g., a credit model may give more weight to
    the accuracy of the predictions for loans with higher dollar amounts. Since
    most evaluation metrics are arithmetic averages, the addition of weights to
    each data point turns the calculations into weighted averages."""

    def __getitem__(self, key: str) -> Any:
        return getattr(self, key)

    def __iter__(self) -> Iterator[str]:
        return (f.name for f in fields(self) if getattr(self, f.name) is not None)

    def __len__(self) -> int:
        return sum(1 for _ in self)

    @property
    def columns(self) -> List[Column]:
        return [v for v in self.values() if isinstance(v, Column)]

    def __call__(self, df: pd.DataFrame) -> Dict[str, Any]:
        return {k: v(df) if isinstance(v, Column) else v for k, v in self.items()}


def _dummy_eval(*args: Any, **kwargs: Any) -> float:
    return np.nan


@dataclass(frozen=True)
class EvaluationMetric(Metric, ABC):
    actual: Column = Column()
    predicted: Column = Column()
    eval: Callable[..., float] = _dummy_eval
    parameters: EvaluationMetricKeywordParameters = field(
        default_factory=EvaluationMetricKeywordParameters,
    )

    def __post_init__(self) -> None:
        # Ignore (i.e. remove) parameters not compatible with the eval
        # function. Note that some metrics, e.g. max_error, doesn't have
        # sample_weight as a parameter, but that's because it doesn't change
        # the result, not because of any inherent incompatibility in meaning,
        # i.e. samples can still be weighted, but max_error will be the same.
        if (
            invalid_parameter_keys := self.parameters.keys()
            - inspect.signature(self.eval).parameters.keys()
        ):
            object.__setattr__(
                self,
                "parameters",
                replace(
                    self.parameters,
                    **dict(
                        zip(
                            invalid_parameter_keys,
                            repeat(None),
                        )
                    ),
                ),
            )

    def operands(self) -> List[Column]:
        return [self.actual, self.predicted] + self.parameters.columns

    def calc(self, df: pd.DataFrame) -> float:
        return self.eval(
            self.actual(df),
            self.predicted(df),
            **self.parameters(df),
        )


@dataclass(frozen=True)
class DriftOperator(UnaryOperator, ABC):
    reference_data: pd.DataFrame = field(
        default_factory=pd.DataFrame,
    )


Distribution: TypeAlias = "pd.Series[float]"
Histogram: TypeAlias = "pd.Series[int]"


@dataclass(frozen=True)
class Discretizer(ABC):
    """Ways to construct histograms from data. Numeric data are commonly
    grouped into intervals while discrete data are grouped into categories.
    This procedure is referred to as binning. Once binned, counts/frequencies
    are tabulated by group to create a histogram.
    """

    binning_method: BinningMethod = CategoricalBinning()

    def histogram(self, data: "pd.Series[Any]") -> Histogram:
        return self.binning_method.histogram(data)


@dataclass(frozen=True)
class DiscreteDivergence(Discretizer, DriftOperator, ABC):
    """See https://en.wikipedia.org/wiki/Divergence_(statistics%29"""

    normalize: Normalizer = AdditiveSmoothing(pseudocount=1)
    """Converts frequencies to probabilities (i.e. normalized to 1)."""

    @abstractmethod
    def divergence(self, pk: Distribution, qk: Distribution) -> float:
        """
        Parameters
        ----------
        pk: series, shape = (d_categories,)
            (discrete) distribution of primary data
        qk: series, shape = (d_categories,)
            (discrete) distribution of reference data,
            a.k.a. the prior distribution

        Returns
        -------
        divergence: float
            divergence of pk over qk
        """

    @cached_property
    def reference_histogram(self) -> Histogram:
        data = self.operand(self.reference_data)
        return self.histogram(data).rename("reference_histogram")

    def calc(self, dataframe: pd.DataFrame) -> float:
        data = self.operand(dataframe)
        # outer-join histograms and fill in zeros for missing categories
        merged_counts = pd.merge(
            self.histogram(data).rename("primary_histogram"),
            self.reference_histogram,
            left_index=True,
            right_index=True,
            how="outer",
            copy=False,
        ).fillna(0)
        # remove rows with all zeros
        merged_counts = merged_counts.loc[(merged_counts > 0).any(axis=1)]
        primary_histogram = merged_counts.primary_histogram
        reference_histogram = merged_counts.reference_histogram
        return self.divergence(
            self.normalize(primary_histogram),
            self.normalize(reference_histogram),
        )



import base64
import dataclasses
from typing import Tuple, Union

import strawberry
from graphql import GraphQLID
from strawberry.custom_scalar import ScalarDefinition
from strawberry.schema.types.scalar import DEFAULT_SCALAR_REGISTRY


def to_global_id(type_name: str, node_id: int) -> str:
    """
    Encode the given id into a global id.

    :param type_name: The type of the node.
    :param node_id: The id of the node.
    :return: A global id.
    """
    return base64.b64encode(f"{type_name}:{node_id}".encode("utf-8")).decode()


def from_global_id(global_id: str) -> Tuple[str, int]:
    """
    Decode the given global id into a type and id.

    :param global_id: The global id to decode.
    :return: A tuple of type and id.
    """
    type_name, node_id = base64.b64decode(global_id).decode().split(":")
    return type_name, int(node_id)


class GlobalIDValueError(ValueError):
    """GlobalID value error, usually related to parsing or serialization."""


@dataclasses.dataclass(frozen=True)
class GlobalID:
    """Global ID for relay types.
    Different from `strawberry.ID`, this ID wraps the original object ID in a string
    that contains both its GraphQL type name and the ID itself, and encodes it
    to a base64_ string.
    This object contains helpers to work with that, including method to retrieve
    the python object type or even the encoded node itself.
    Attributes:
        type_name:
            The type name part of the id
        node_id:
            The node id part of the id
    .. _base64:
        https://en.wikipedia.org/wiki/Base64
    """

    type_name: str
    node_id: int

    def __post_init__(self) -> None:
        if not isinstance(self.type_name, str):
            raise GlobalIDValueError(
                f"type_name is expected to be a string, found {self.type_name}"
            )
        try:
            # node_id could be numpy.int64, hence the need for coercion
            object.__setattr__(self, "node_id", int(self.node_id))
        except ValueError:
            raise GlobalIDValueError(f"node_id is expected to be an int, found {self.node_id}")

    def __str__(self) -> str:
        return to_global_id(self.type_name, self.node_id)

    @classmethod
    def from_id(cls, value: Union[str, strawberry.ID]) -> "GlobalID":
        """Create a new GlobalID from parsing the given value.
        Args:
            value:
                The value to be parsed, as a base64 string in the "TypeName:NodeID" format
        Returns:
            An instance of GLobalID
        Raises:
            GlobalIDValueError:
                If the value is not in a GLobalID format
        """
        try:
            type_name, node_id = from_global_id(value)
        except ValueError as e:
            raise GlobalIDValueError(str(e)) from e

        return cls(type_name=type_name, node_id=node_id)


@strawberry.interface(description="A node in the graph with a globally unique ID")
class Node:
    """
    All types that are relay ready should inherit from this interface and
    implement the following methods.

    Attributes:
        id_attr:
            The raw id field of node. Typically a database id or index
    """

    id_attr: strawberry.Private[int]

    @strawberry.field
    def id(self) -> GlobalID:
        return GlobalID(type(self).__name__, self.id_attr)


# Register our GlobalID scalar
DEFAULT_SCALAR_REGISTRY[GlobalID] = ScalarDefinition(
    # Use the same name/description/parse_literal from GraphQLID
    # specs expect this type to be "ID".
    name="GlobalID",
    description=GraphQLID.description,
    parse_literal=lambda v, vars=None: GlobalID.from_id(GraphQLID.parse_literal(v, vars)),
    parse_value=GlobalID.from_id,
    serialize=str,
    specified_by_url="https://relay.dev/graphql/objectidentification.htm",
)



import logging
from pathlib import Path
from typing import Optional, Union

from starlette.applications import Starlette
from starlette.datastructures import QueryParams
from starlette.endpoints import HTTPEndpoint
from starlette.exceptions import HTTPException
from starlette.middleware import Middleware
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import FileResponse, Response
from starlette.routing import Mount, Route, WebSocketRoute
from starlette.staticfiles import StaticFiles
from starlette.types import Scope
from starlette.websockets import WebSocket
from strawberry.asgi import GraphQL
from strawberry.schema import BaseSchema

from phoenix.config import SERVER_DIR
from phoenix.core.model_schema import Model

from .api.context import Context
from .api.schema import schema

logger = logging.getLogger(__name__)


class Static(StaticFiles):
    "Static file serving with a fallback to index.html"

    async def get_response(self, path: str, scope: Scope) -> Response:
        response = None
        try:
            response = await super().get_response(path, scope)
        except HTTPException as e:
            if e.status_code != 404:
                raise e
            # Fallback to to the index.html
            full_path, stat_result = self.lookup_path("index.html")
            if stat_result is None:
                raise RuntimeError("Failed to find index.html")
            response = self.file_response(full_path, stat_result, scope)
        except Exception as e:
            raise e
        return response


class HeadersMiddleware(BaseHTTPMiddleware):
    async def dispatch(
        self,
        request: Request,
        call_next: RequestResponseEndpoint,
    ) -> Response:
        response = await call_next(request)
        response.headers["x-colab-notebook-cache-control"] = "no-cache"
        return response


class GraphQLWithContext(GraphQL):  # type: ignore
    def __init__(
        self,
        schema: BaseSchema,
        model: Model,
        export_path: Path,
        graphiql: bool = False,
        corpus: Optional[Model] = None,
    ) -> None:
        self.model = model
        self.corpus = corpus
        self.export_path = export_path
        super().__init__(schema, graphiql=graphiql)

    async def get_context(
        self,
        request: Union[Request, WebSocket],
        response: Optional[Response] = None,
    ) -> Context:
        return Context(
            request=request,
            response=response,
            model=self.model,
            corpus=self.corpus,
            export_path=self.export_path,
        )


class Download(HTTPEndpoint):
    path: Path

    async def get(self, request: Request) -> FileResponse:
        params = QueryParams(request.query_params)
        file = self.path / (params.get("filename", "") + ".parquet")
        if not file.is_file():
            raise HTTPException(status_code=404)
        return FileResponse(
            path=file,
            filename=file.name,
            media_type="application/x-octet-stream",
        )


def create_app(
    export_path: Path,
    model: Model,
    corpus: Optional[Model] = None,
    debug: bool = False,
) -> Starlette:
    graphql = GraphQLWithContext(
        schema=schema,
        model=model,
        corpus=corpus,
        export_path=export_path,
        graphiql=True,
    )
    return Starlette(
        middleware=[
            Middleware(HeadersMiddleware),
        ],
        debug=debug,
        routes=[
            Route(
                "/exports",
                type(
                    "DownloadExports",
                    (Download,),
                    {"path": export_path},
                ),
            ),
            Route(
                "/graphql",
                graphql,
            ),
            WebSocketRoute("/graphql", graphql),
            Mount(
                "/",
                app=Static(
                    directory=SERVER_DIR / "static",
                    html=True,
                ),
                name="static",
            ),
        ],
    )



from enum import Enum
from functools import partial

import strawberry

from phoenix.metrics.wrappers import SkEval


@strawberry.enum
class PerformanceMetric(Enum):
    # To become enum values, functions need to be wrapped in partial.
    # See https://stackoverflow.com/a/40339397
    accuracyScore = partial(SkEval.accuracy_score)  # type: ignore



from datetime import datetime
from typing import Iterable, List, Optional, Set, Union

import strawberry
from strawberry.scalars import ID
from strawberry.unset import UNSET

import phoenix.core.model_schema as ms
from phoenix.core.model_schema import FEATURE, TAG, ScalarDimension

from ..input_types.DimensionInput import DimensionInput
from .DatasetRole import AncillaryDatasetRole, DatasetRole
from .Dimension import Dimension, to_gql_dimension
from .Event import Event, create_event, create_event_id, parse_event_ids_by_dataset_role


@strawberry.type
class Dataset:
    start_time: datetime = strawberry.field(description="The start bookend of the data")
    end_time: datetime = strawberry.field(description="The end bookend of the data")
    dataset: strawberry.Private[ms.Dataset]
    dataset_role: strawberry.Private[Union[DatasetRole, AncillaryDatasetRole]]
    model: strawberry.Private[ms.Model]

    # type ignored here to get around the following: https://github.com/strawberry-graphql/strawberry/issues/1929
    @strawberry.field(description="Returns a human friendly name for the dataset.")  # type: ignore
    def name(self) -> str:
        return self.dataset.display_name

    @strawberry.field
    def events(
        self,
        event_ids: List[ID],
        dimensions: Optional[List[DimensionInput]] = UNSET,
    ) -> List[Event]:
        """
        Returns events for specific event IDs and dimensions. If no input
        dimensions are provided, returns all features and tags.
        """
        if not event_ids:
            return []
        row_ids = parse_event_ids_by_dataset_role(event_ids)
        if len(row_ids) > 1 or self.dataset_role not in row_ids:
            raise ValueError("eventIds contains IDs from incorrect dataset.")
        events = self.dataset[row_ids[self.dataset_role]]
        requested_gql_dimensions = _get_requested_features_and_tags(
            core_dimensions=self.model.scalar_dimensions,
            requested_dimension_names=set(dim.name for dim in dimensions)
            if isinstance(dimensions, list)
            else None,
        )
        return [
            create_event(
                event_id=create_event_id(event.id.row_id, self.dataset_role),
                event=event,
                dimensions=requested_gql_dimensions,
                is_document_record=self.dataset_role is AncillaryDatasetRole.corpus,
            )
            for event in events
        ]


def _get_requested_features_and_tags(
    core_dimensions: Iterable[ScalarDimension],
    requested_dimension_names: Optional[Set[str]] = UNSET,
) -> List[Dimension]:
    """
    Returns requested features and tags as a list of strawberry Datasets. If no
    dimensions are explicitly requested, returns all features and tags.
    """
    requested_features_and_tags: List[Dimension] = []
    for id, dim in enumerate(core_dimensions):
        is_requested = (
            not isinstance(requested_dimension_names, Set)
        ) or dim.name in requested_dimension_names
        is_feature_or_tag = dim.role in (FEATURE, TAG)
        if is_requested and is_feature_or_tag:
            requested_features_and_tags.append(to_gql_dimension(id_attr=id, dimension=dim))
    return requested_features_and_tags



from typing import Optional

import strawberry

from phoenix.server.api.interceptor import GqlValueMediator


@strawberry.type
class PromptResponse:
    """
    A prompt and response pair. Used to group together the input and output of a LLM
    """

    prompt: Optional[str] = strawberry.field(
        description="The prompt submitted to the LLM",
        default=GqlValueMediator(),
    )
    response: Optional[str] = strawberry.field(
        description="The response generated by the LLM",
        default=GqlValueMediator(),
    )

    def __bool__(self) -> bool:
        return bool(self.prompt) or bool(self.response)



from typing import Optional

import strawberry

from phoenix.server.api.interceptor import GqlValueMediator


@strawberry.type
class EventMetadata:
    """A partial record of a model event. E.g. the predictions and actuals"""

    prediction_id: Optional[str] = strawberry.field(default=GqlValueMediator())
    prediction_score: Optional[float] = strawberry.field(default=GqlValueMediator())
    prediction_label: Optional[str] = strawberry.field(default=GqlValueMediator())
    actual_score: Optional[float] = strawberry.field(default=GqlValueMediator())
    actual_label: Optional[str] = strawberry.field(default=GqlValueMediator())



import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Iterable, List, Mapping, Optional, Union

import numpy as np
import pandas as pd

from phoenix.core.model_schema import Column

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Metric(ABC):
    def id(self) -> int:
        """
        id is a unique identifier for each metric instance. This is used to
        extract the metric's own value from a collective output containing
        results from other metrics.
        """
        return id(self)

    @property
    def initial_value(self) -> Any:
        return np.nan

    def get_value(self, result: Mapping[int, Any]) -> Any:
        try:
            return result[self.id()]
        except KeyError:
            return self.initial_value

    @abstractmethod
    def calc(self, dataframe: pd.DataFrame) -> Any:
        ...

    @abstractmethod
    def operands(self) -> List[Column]:
        ...

    def __call__(
        self,
        df: pd.DataFrame,
        /,
        subset_rows: Optional[Union[slice, List[int]]] = None,
    ) -> Any:
        """
        Computes the metric on a dataframe.

        Parameters
        ----------
        df: pandas DataFrame
            The dataframe input to the metric.
        subset_rows: Optional[Union[slice, List[int]]] = None
            Optionally specifying a subset of rows for the computation.
            Can be a list or slice (e.g. `slice(100, 200)`) of integers.
        """
        subset_rows = slice(None) if subset_rows is None else subset_rows
        df = df.iloc[
            subset_rows,
            sorted(
                {
                    df.columns.get_loc(name)
                    for name in map(str, self.operands())
                    if name in df.columns
                }
            ),
        ]
        try:
            return self.calc(df)
        except (TypeError, ValueError, NotImplementedError) as exc:
            logger.warning(exc, exc_info=True)
            return self.initial_value


def multi_calculate(
    df: pd.DataFrame,
    calcs: Iterable[Metric],
) -> "pd.Series[Any]":
    """
    Calculates multiple metrics on the same dataframe.
    """
    return pd.Series({calc.id(): calc(df) for calc in calcs})



from dataclasses import dataclass
from typing import Dict, List, Mapping, Protocol, Set, Tuple

import numpy as np
import numpy.typing as npt
from strawberry import ID
from typing_extensions import TypeAlias

from phoenix.pointcloud.clustering import RawCluster

Vector: TypeAlias = npt.NDArray[np.float64]
Matrix: TypeAlias = npt.NDArray[np.float64]
RowIndex: TypeAlias = int


class DimensionalityReducer(Protocol):
    def project(self, mat: Matrix, n_components: int) -> Matrix:
        ...


class ClustersFinder(Protocol):
    def find_clusters(self, mat: Matrix) -> List[RawCluster]:
        ...


@dataclass(frozen=True)
class PointCloud:
    dimensionalityReducer: DimensionalityReducer
    clustersFinder: ClustersFinder

    def generate(
        self,
        data: Mapping[ID, Vector],
        n_components: int = 3,
    ) -> Tuple[Dict[ID, Vector], Dict[str, Set[ID]]]:
        """
        Given a set of vectors, projects them onto lower dimensions, and
        finds clusters among the projections.

        Parameters
        ----------
        data : mapping
            Mapping of input vectors by their EventIds.

        n_components: int, default=3
            Number of dimensions in the projected space.

        Returns
        -------
        projections : dictionary
            Projected vectors in the low dimensional space, mapped back to the
            input vectors' EventIds.

        cluster_membership: dictionary
            Cluster membership by way of cluster_ids in the form of integers
            0,1,2,... mapped back to the input vectors' EventIds. Note that
            some vectors may not belong to any cluster and are excluded here.

        """

        if not data:
            return {}, {}
        event_ids, vectors = zip(*data.items())
        projections = self.dimensionalityReducer.project(
            np.stack(vectors), n_components=n_components
        )
        clusters = self.clustersFinder.find_clusters(projections)
        return dict(zip(event_ids, projections)), {
            str(i): {event_ids[row_index] for row_index in cluster}
            for i, cluster in enumerate(clusters)
        }



from collections import Counter, defaultdict
from typing import Dict, List, Mapping, Optional, Set

import strawberry
from strawberry import ID
from strawberry.types import Info

from phoenix.core.model_schema import PRIMARY, REFERENCE
from phoenix.server.api.context import Context
from phoenix.server.api.input_types.DataQualityMetricInput import DataQualityMetricInput
from phoenix.server.api.input_types.PerformanceMetricInput import PerformanceMetricInput
from phoenix.server.api.types.DatasetRole import AncillaryDatasetRole, DatasetRole
from phoenix.server.api.types.DatasetValues import DatasetValues
from phoenix.server.api.types.Event import unpack_event_id


@strawberry.type
class Cluster:
    """A grouping of points in a UMAP plot"""

    id: ID = strawberry.field(
        description="The ID of the cluster",
    )

    event_ids: List[ID] = strawberry.field(
        description="The event IDs of the points in the cluster",
    )

    @strawberry.field(
        description="Ratio of primary points over reference points",
    )  # type: ignore
    def drift_ratio(
        self,
        info: Info[Context, None],
    ) -> Optional[float]:
        """
        Calculates the drift score of the cluster. The score will be a value
        representing the balance of points between the primary and the reference
        datasets, and will be on a scale between 1 (all primary) and -1 (all
        reference), with 0 being an even balance between the two datasets.

        Returns
        -------
        drift_ratio : Optional[float]
        """
        model = info.context.model
        if model[REFERENCE].empty:
            return None
        count_by_role = Counter(unpack_event_id(event_id)[1] for event_id in self.event_ids)
        primary_count = count_by_role[DatasetRole.primary]
        reference_count = count_by_role[DatasetRole.reference]
        return (
            None
            if not (denominator := (primary_count + reference_count))
            else (primary_count - reference_count) / denominator
        )

    @strawberry.field(
        description="Ratio of primary points over corpus points",
    )  # type: ignore
    def primary_to_corpus_ratio(
        self,
        info: Info[Context, None],
    ) -> Optional[float]:
        """
        Calculates a score representing the balance of points between the
        primary and the corpus datasets, and will be on a scale between 1
        (all primary) and -1 (all corpus), with 0 being an even balance
        between the two datasets.

        Returns
        -------
        drift_ratio : Optional[float]
        """
        corpus = info.context.corpus
        if corpus is None or corpus[PRIMARY].empty:
            return None
        count_by_role = Counter(unpack_event_id(event_id)[1] for event_id in self.event_ids)
        primary_count = count_by_role[DatasetRole.primary]
        corpus_count = count_by_role[AncillaryDatasetRole.corpus]
        return (
            None
            if not (denominator := (primary_count + corpus_count))
            else (primary_count - corpus_count) / denominator
        )

    @strawberry.field(
        description="Data quality metric summarized by the respective "
        "datasets of the clustered events",
    )  # type: ignore
    def data_quality_metric(
        self,
        info: Info[Context, None],
        metric: DataQualityMetricInput,
    ) -> DatasetValues:
        model = info.context.model
        row_ids: Dict[DatasetRole, List[int]] = defaultdict(list)
        for row_id, dataset_role in map(unpack_event_id, self.event_ids):
            if not isinstance(dataset_role, DatasetRole):
                continue
            row_ids[dataset_role].append(row_id)
        return DatasetValues(
            primary_value=metric.metric_instance(
                model[PRIMARY],
                subset_rows=row_ids[DatasetRole.primary],
            ),
            reference_value=metric.metric_instance(
                model[REFERENCE],
                subset_rows=row_ids[DatasetRole.reference],
            ),
        )

    @strawberry.field(
        description="Performance metric summarized by the respective "
        "datasets of the clustered events",
    )  # type: ignore
    def performance_metric(
        self,
        info: Info[Context, None],
        metric: PerformanceMetricInput,
    ) -> DatasetValues:
        model = info.context.model
        row_ids: Dict[DatasetRole, List[int]] = defaultdict(list)
        for row_id, dataset_role in map(unpack_event_id, self.event_ids):
            if not isinstance(dataset_role, DatasetRole):
                continue
            row_ids[dataset_role].append(row_id)
        metric_instance = metric.metric_instance(model)
        return DatasetValues(
            primary_value=metric_instance(
                model[PRIMARY],
                subset_rows=row_ids[DatasetRole.primary],
            ),
            reference_value=metric_instance(
                model[REFERENCE],
                subset_rows=row_ids[DatasetRole.reference],
            ),
        )


def to_gql_clusters(
    clustered_events: Mapping[str, Set[ID]],
) -> List[Cluster]:
    """
    Converts a dictionary of event IDs to cluster IDs to a list of clusters
    for the graphQL response

    Parameters
    ----------
    clustered_events: Mapping[str, Set[ID]]
        A mapping of cluster ID to its set of event IDs
    """

    return [
        Cluster(
            id=ID(cluster_id),
            event_ids=list(event_ids),
        )
        for cluster_id, event_ids in clustered_events.items()
    ]



import json
import logging
from dataclasses import dataclass, replace
from enum import Enum, auto
from pathlib import Path
from typing import Iterator, NamedTuple, Optional, Tuple
from urllib import request
from urllib.parse import quote, urljoin

from pandas import read_parquet

from phoenix.config import DATASET_DIR
from phoenix.datasets.dataset import Dataset
from phoenix.datasets.schema import (
    EmbeddingColumnNames,
    RetrievalEmbeddingColumnNames,
    Schema,
)

logger = logging.getLogger(__name__)


class DatasetRole(Enum):
    PRIMARY = auto()
    REFERENCE = auto()
    CORPUS = auto()


@dataclass(frozen=True)
class Fixture:
    name: str
    description: str
    prefix: str
    primary_file_name: str
    primary_schema: Schema
    reference_file_name: Optional[str] = None
    # The schema for the reference dataset. If not provided, the primary schema will be used
    reference_schema: Optional[Schema] = None
    corpus_file_name: Optional[str] = None
    corpus_schema: Optional[Schema] = None

    def paths(self) -> Iterator[Tuple[DatasetRole, Path]]:
        return (
            (role, Path(self.prefix) / name)
            for role, name in zip(
                DatasetRole,
                (
                    self.primary_file_name,
                    self.reference_file_name,
                    self.corpus_file_name,
                ),
            )
            if name
        )


sentiment_classification_language_drift_schema = Schema(
    prediction_id_column_name="prediction_id",
    timestamp_column_name="prediction_ts",
    prediction_label_column_name="pred_label",
    actual_label_column_name="label",
    embedding_feature_column_names={
        "text_embedding": EmbeddingColumnNames(
            vector_column_name="text_vector", raw_data_column_name="text"
        ),
    },
)
sentiment_classification_language_drift_fixture = Fixture(
    name="sentiment_classification_language_drift",
    description="""
    Highlights issues that occur maintaining a sentiment classification model.
    This  model takes online reviews of your U.S.-based product as the input and
    predicts whether the reviewer's sentiment was positive, negative, or
    neutral.

    You trained your sentiment classification model on English reviews. However,
    once the model was released into production, you notice that the performance
    of the model has degraded over a period of time.

    Phoenix is able to surface the reason for this performance degradation. In
    this example, the presence of reviews written in Spanish impact the model's
    performance. You can surface and troubleshoot this issue by analyzing the
    embedding vectors associated with the online review text.
    """,
    primary_schema=sentiment_classification_language_drift_schema,
    reference_schema=sentiment_classification_language_drift_schema,
    prefix="unstructured/nlp/sentiment-classification-language-drift",
    primary_file_name="sentiment_classification_language_drift_production.parquet",
    reference_file_name="sentiment_classification_language_drift_training.parquet",
)

image_classification_schema = Schema(
    timestamp_column_name="prediction_ts",
    prediction_label_column_name="predicted_action",
    actual_label_column_name="actual_action",
    embedding_feature_column_names={
        "image_embedding": EmbeddingColumnNames(
            vector_column_name="image_vector",
            link_to_data_column_name="url",
        ),
    },
)

image_classification_fixture = Fixture(
    name="image_classification",
    description="""
    Imagine you're in charge of maintaining a model that classifies the action
    of people in photographs. Your model initially performs well in production,
    but its performance gradually degrades over time.
    """,
    primary_schema=replace(image_classification_schema, actual_label_column_name=None),
    reference_schema=image_classification_schema,
    prefix="unstructured/cv/human-actions",
    primary_file_name="human_actions_production.parquet",
    reference_file_name="human_actions_training.parquet",
)

fashion_mnist_primary_schema = Schema(
    timestamp_column_name="prediction_ts",
    embedding_feature_column_names={
        "embedding": EmbeddingColumnNames(
            vector_column_name="embeddings", link_to_data_column_name="image_url"
        ),
    },
    actual_label_column_name="actual_label",
    prediction_label_column_name="predicted_label",
)
fashion_mnist_reference_schema = replace(fashion_mnist_primary_schema, timestamp_column_name=None)
fashion_mnist_fixture = Fixture(
    name="fashion_mnist",
    description="""
    Fashion-MNIST is a dataset of Zalando's article images consisting of a
    training set of 60,000 examples and a test set of 10,000 examples. Each
    example is a 28x28 grayscale image, associated with a label from 10 classes.
    Fashion-MNIST serves as a direct drop-in replacement for the original MNIST
    dataset for benchmarking machine learning algorithms. It shares the same
    image size and structure of training and testing splits.
    """,
    primary_schema=fashion_mnist_primary_schema,
    reference_schema=fashion_mnist_reference_schema,
    prefix="unstructured/cv/fashion-mnist",
    primary_file_name="fashion_mnist_production.parquet",
    reference_file_name="fashion_mnist_train.parquet",
)

ner_token_drift_schema = Schema(
    timestamp_column_name="prediction_ts",
    feature_column_names=["language"],
    actual_label_column_name="label",
    prediction_label_column_name="pred_label",
    embedding_feature_column_names={
        "embedding": EmbeddingColumnNames(
            vector_column_name="token_vector", raw_data_column_name="text"
        )
    },
)
ner_token_drift_fixture = Fixture(
    name="ner_token_drift",
    description="""
    You are in charge of maintaining a Named Entity Recognition (NER) model.
    This simple model can automatically scan text, pull out some fundamental
    entities within it, and classify them into predefined categories: Person,
    Location, or Organization. However, once the model was released into
    production, you notice that the performance of the model has degraded over a
    period of time.

    Phoenix is able to surface the reason for this performance degradation. In
    this example, text including locations is under-represented in the training
    set. This label imbalance impacts the model's performance. You can surface
    and troubleshoot this issue by analyzing the embedding vectors associated
    with the input text.
    """,
    primary_schema=ner_token_drift_schema,
    reference_schema=ner_token_drift_schema,
    prefix="unstructured/nlp/named-entity-recognition",
    primary_file_name="ner_token_drift_production.parquet",
    reference_file_name="ner_token_drift_train.parquet",
)

credit_card_fraud_schema = Schema(
    prediction_id_column_name="prediction_id",
    prediction_label_column_name="predicted_label",
    prediction_score_column_name="predicted_score",
    actual_label_column_name="actual_label",
    timestamp_column_name="prediction_timestamp",
    tag_column_names=["age"],
    embedding_feature_column_names={
        "tabular_embedding": EmbeddingColumnNames(vector_column_name="tabular_vector"),
    },
)
credit_card_fraud_fixture = Fixture(
    name="credit_card_fraud",
    description="""
    Use-case for a credit card fraud detection model at a large bank or payment
    processing company.

    You have been alerted by a spike in credit card chargebacks leading you to
    suspect that fraudsters are getting away with committing fraud undetected!

    Realizing that this flaw in your model's performance has a heavy cost on
    your company and customers, you understand the need for a powerful tools to
    troubleshoot and prevent costly model degradations. You turn to Phoenix to
    find out what changed in your credit card fraud detection model and how you
    can improve it.
    """,
    primary_schema=credit_card_fraud_schema,
    reference_schema=credit_card_fraud_schema,
    prefix="structured/credit-card-fraud",
    primary_file_name="credit_card_fraud_production.parquet",
    reference_file_name="credit_card_fraud_train.parquet",
)

click_through_rate_schema = Schema(
    timestamp_column_name="prediction_timestamp",
    prediction_id_column_name="prediction_id",
    prediction_label_column_name="predicted_label",
    prediction_score_column_name="predicted_score",
    actual_label_column_name="actual_label",
)
click_through_rate_fixture = Fixture(
    name="click_through_rate",
    description="""
    Investigate various performance related aspects of an online advertisement
    use-case. These datasets are designed for analyzing Click-through Rate (CTR)
    performance.

    You manage the models for an online advertising platform. You have spent a
    great deal of your time collecting online data and training models for best
    performance. With your models now in production you have no tools available
    to your disposal to monitor the performance of your models, identify any
    issues, or get insights into how to improve your models.

    This use-case highlights a common advertisement use-case and is tailored for
    analyzing CTR for an ad or ad group.
    """,
    primary_schema=click_through_rate_schema,
    reference_schema=click_through_rate_schema,
    prefix="structured/click-through-rate",
    primary_file_name="click_through_rate_production.parquet",
    reference_file_name="click_through_rate_train.parquet",
)

wide_data_primary_schema = Schema(
    actual_label_column_name="actual_label",
    prediction_label_column_name="predicted_label",
    timestamp_column_name="prediction_ts",
)
wide_data_reference_schema = replace(wide_data_primary_schema, timestamp_column_name=None)
wide_data_fixture = Fixture(
    name="wide_data",
    description="""
    Use-case that for a wide data model (e.g. a large amount of features). For
    developer use only.
    """,
    primary_schema=wide_data_primary_schema,
    reference_schema=wide_data_reference_schema,
    prefix="structured/wide-data",
    primary_file_name="wide_data_production.parquet",
    reference_file_name="wide_data_train.parquet",
)

deep_data_primary_schema = Schema(
    timestamp_column_name="prediction_ts",
    actual_label_column_name="actual_label",
    prediction_label_column_name="predicted_label",
)
deep_data_reference_schema = replace(deep_data_primary_schema, timestamp_column_name=None)
deep_data_fixture = Fixture(
    name="deep_data",
    description="""
    Use-case that for a deep data model (e.g. a lot of feature values).
    For developer use only.
    """,
    primary_schema=deep_data_primary_schema,
    reference_schema=deep_data_reference_schema,
    prefix="structured/deep-data",
    primary_file_name="deep_data_production.parquet",
    reference_file_name="deep_data_train.parquet",
)


llm_summarization_schema = Schema(
    timestamp_column_name="prediction_timestamp",
    tag_column_names=[
        "rougeL_score",
        "reference_summary",
    ],
    prompt_column_names=EmbeddingColumnNames(
        vector_column_name="article_vector", raw_data_column_name="article"
    ),
    response_column_names=EmbeddingColumnNames(
        vector_column_name="summary_vector", raw_data_column_name="summary"
    ),
)
llm_summarization_fixture = Fixture(
    name="llm_summarization",
    description="""
    LLM summarization data.
    """,
    primary_schema=llm_summarization_schema,
    reference_schema=llm_summarization_schema,
    prefix="unstructured/llm/summarization",
    primary_file_name="llm_summarization_prod.parquet",
    reference_file_name="llm_summarization_baseline.parquet",
)

wikipedia_fixture = Fixture(
    name="wiki",
    description="""
    Semantic search dataset including queries, answers, retrievals and corpus
    documents. Queries are sampled from Google’s Natural Questions dataset
    https://ai.google.com/research/NaturalQuestions, and documents (paragraphs)
    are sampled from http://sbert.net/datasets/simplewiki-2020-11-01.jsonl.gz,
    based on the Simple English Wikipedia. Embeddings for both questions and
    documents are generated from the bi-encoder 'multi-qa-MiniLM-L6-cos-v1',
    which produces normalized vectors that are 384-dimensional. Given a query
    and its embedding, a small subset of documents are first retrieved based on
    cosine similarity search on the document embeddings, then relevance scores
    are computed on the retrieved documents using the cross-encoder
    'cross-encoder/ms-marco-MiniLM-L-12-v2'. Answers (to the queries) are
    generated from 'deepset/tinyroberta-squad2' using the most relevant
    documents as the context input.

    References:
    - https://www.sbert.net/examples/applications/semantic-search/README.html
    - https://github.com/UKPLab/sentence-transformers/tree/master/examples/applications/retrieve_rerank
    - https://www.sbert.net/docs/pretrained_models.html
    - https://www.sbert.net/docs/pretrained_cross-encoders.html
    - https://ai.google.com/research/NaturalQuestions
    """,  # noqa: E501
    primary_schema=Schema(
        prediction_id_column_name="id",
        prompt_column_names=RetrievalEmbeddingColumnNames(
            vector_column_name="embedding",
            raw_data_column_name="question",
            context_retrieval_ids_column_name="retrievals",
            context_retrieval_scores_column_name="scores",
        ),
        response_column_names="answer",
    ),
    corpus_schema=Schema(
        prediction_id_column_name="id",
        prompt_column_names=EmbeddingColumnNames(
            vector_column_name="embedding",
            raw_data_column_name="text",
        ),
    ),
    prefix="unstructured/search/wiki",
    primary_file_name="queries.parquet",
    # The reference file is intended purely for troubleshooting bad data. Un-comment for testing.
    # reference_file_name="queries2.parquet",
    corpus_file_name="corpus.parquet",
)

FIXTURES: Tuple[Fixture, ...] = (
    sentiment_classification_language_drift_fixture,
    image_classification_fixture,
    fashion_mnist_fixture,
    ner_token_drift_fixture,
    credit_card_fraud_fixture,
    click_through_rate_fixture,
    wide_data_fixture,
    deep_data_fixture,
    llm_summarization_fixture,
    wikipedia_fixture,
)
NAME_TO_FIXTURE = {fixture.name: fixture for fixture in FIXTURES}


def get_datasets(
    fixture_name: str,
    no_internet: bool = False,
) -> Tuple[Dataset, Optional[Dataset], Optional[Dataset]]:
    """
    Downloads primary and reference datasets for a fixture if they are not found
    locally.
    """
    fixture = _get_fixture_by_name(fixture_name=fixture_name)
    if no_internet:
        paths = {role: DATASET_DIR / path for role, path in fixture.paths()}
    else:
        paths = dict(_download(fixture, DATASET_DIR))
    primary_dataset = Dataset(
        read_parquet(paths[DatasetRole.PRIMARY]),
        fixture.primary_schema,
        "production",
    )
    reference_dataset = None
    if fixture.reference_file_name is not None:
        reference_dataset = Dataset(
            read_parquet(paths[DatasetRole.REFERENCE]),
            fixture.reference_schema
            if fixture.reference_schema is not None
            else fixture.primary_schema,
            "training",
        )
    corpus_dataset = None
    if fixture.corpus_file_name is not None:
        corpus_dataset = Dataset(
            read_parquet(paths[DatasetRole.CORPUS]),
            fixture.corpus_schema,
            "knowledge_base",
        )
    return primary_dataset, reference_dataset, corpus_dataset


def _get_fixture_by_name(fixture_name: str) -> Fixture:
    """
    Returns the fixture whose name matches the input name. Raises a ValueError
    if the input fixture name does not match any known fixture names.
    """
    if fixture_name not in NAME_TO_FIXTURE:
        valid_fixture_names = ", ".join(NAME_TO_FIXTURE.keys())
        raise ValueError(f'"{fixture_name}" is invalid. Valid names are: {valid_fixture_names}')
    return NAME_TO_FIXTURE[fixture_name]


@dataclass
class ExampleDatasets:
    """
    A primary and optional reference dataset pair.
    """

    primary: Dataset
    reference: Optional[Dataset] = None
    corpus: Optional[Dataset] = None


def load_example(use_case: str) -> ExampleDatasets:
    """
    Loads an example primary and reference dataset for a given use-case.

    Parameters
    ----------
        use_case: str
            Name of the phoenix supported use case Valid values include:
                - "sentiment_classification_language_drift"
                - "image_classification"
                - "fashion_mnist"
                - "ner_token_drift"
                - "credit_card_fraud"
                - "click_through_rate"


    Returns
    _______
        datasets: DatasetDict
            A dictionary of datasets, split out by dataset type (primary,
            reference).

    """
    fixture = _get_fixture_by_name(use_case)
    primary_dataset, reference_dataset, corpus_dataset = get_datasets(use_case)
    print(f"📥 Loaded {use_case} example datasets.")
    print("ℹ️ About this use-case:")
    print(fixture.description)
    return ExampleDatasets(
        primary=primary_dataset,
        reference=reference_dataset,
        corpus=corpus_dataset,
    )


class Metadata(NamedTuple):
    path: str
    mediaLink: str
    md5Hash: str

    def save_artifact(self, location: Path) -> Path:
        data_file_path = location / self.path
        md5_file = data_file_path.with_name(data_file_path.stem + ".md5")
        data_file_path.parents[0].mkdir(parents=True, exist_ok=True)
        if data_file_path.is_file() and md5_file.is_file():
            with open(md5_file, "r") as f:
                if f.readline() == self.md5Hash:
                    return data_file_path
        request.urlretrieve(self.mediaLink, data_file_path)
        with open(md5_file, "w") as f:
            f.write(self.md5Hash)
        return data_file_path


class GCSAssets(NamedTuple):
    host: str = "https://storage.googleapis.com/"
    bucket: str = "arize-assets"
    prefix: str = "phoenix/datasets/"

    def metadata(self, path: Path) -> Metadata:
        url = urljoin(
            urljoin(self.host, f"storage/v1/b/{self.bucket}/o/"),
            quote(urljoin(self.prefix, str(path)), safe=""),
        )
        resp = json.loads(request.urlopen(request.Request(url)).read())
        return Metadata(
            resp["name"][len(self.prefix) :],
            resp["mediaLink"],
            resp["md5Hash"],
        )


def _download(fixture: Fixture, location: Path) -> Iterator[Tuple[DatasetRole, Path]]:
    for role, path in fixture.paths():
        yield role, GCSAssets().metadata(path).save_artifact(location)


# Download all fixtures
if __name__ == "__main__":
    import time

    for fixture in FIXTURES:
        start_time = time.time()
        print(f"getting {fixture.name}", end="...")
        dict(_download(fixture, DATASET_DIR))
        print(f"done ({time.time() - start_time:.2f}s)")



from datetime import datetime, timedelta, timezone
from functools import partial
from itertools import accumulate, repeat
from typing import Callable, Iterable, Iterator, Tuple, cast

import pandas as pd
from typing_extensions import TypeAlias

from phoenix.metrics import Metric, multi_calculate


def timeseries(
    *,
    start_time: datetime,
    end_time: datetime,
    evaluation_window: timedelta,
    sampling_interval: timedelta,
) -> Callable[..., pd.DataFrame]:
    """
    Returns an aggregator for use by pandas.DataFrame.pipe() in conjunction
    with a list of Metrics to return a final time-series output in the form
    of a pandas.DataFrame where the row index is timestamp and each column
    is identified by each Metric's ID. Apply each metric's get_value() on
    the rows to extract the corresponding metric output value.
    """

    return partial(
        _aggregator,
        start_time=start_time,
        end_time=end_time,
        evaluation_window=evaluation_window,
        sampling_interval=sampling_interval,
    )


StartIndex: TypeAlias = int
StopIndex: TypeAlias = int


def row_interval_from_sorted_time_index(
    time_index: pd.Index,
    time_start: datetime,
    time_stop: datetime,
) -> Tuple[StartIndex, StopIndex]:
    """
    Returns end exclusive time slice from sorted index.
    """
    return cast(
        Tuple[StartIndex, StopIndex],
        time_index.searchsorted((time_start, time_stop)),
    )


def _aggregator(
    dataframe: pd.DataFrame,
    *,
    metrics: Iterable[Metric],
    start_time: datetime,
    end_time: datetime,
    evaluation_window: timedelta,
    sampling_interval: timedelta,
) -> pd.DataFrame:
    """
    Calls groupby on the dataframe and apply metric calculations on each group.
    """
    calcs = tuple(metrics)
    return pd.concat(
        _results(
            calcs=calcs,
            dataframe=dataframe,
            start_time=start_time,
            end_time=end_time,
            evaluation_window=evaluation_window,
            sampling_interval=sampling_interval,
        ),
        verify_integrity=True,
    )


StartTime: TypeAlias = datetime
EndTime: TypeAlias = datetime


def _groupers(
    start_time: datetime,
    end_time: datetime,
    evaluation_window: timedelta,
    sampling_interval: timedelta,
) -> Iterator[Tuple[StartTime, EndTime, pd.Grouper]]:
    """
    Yields pandas.Groupers from time series parameters.
    """
    if not sampling_interval:
        return
    total_time_span = end_time - start_time
    divisible = evaluation_window % sampling_interval == timedelta()
    if divisible and evaluation_window < total_time_span:
        max_offset = evaluation_window
    else:
        max_offset = total_time_span
    offsets = accumulate(
        repeat(sampling_interval),
        initial=timedelta(),
    )
    for offset in offsets:
        if offset >= max_offset:
            return
        # Each Grouper is like a row in a brick wall, where each brick is an
        # evaluation window. By shifting each row of bricks by the sampling
        # interval, we can get all the brick's right edges to line up with the
        # points of the time series, and together they will summarize data for
        # the whole time series.
        #
        #                   evaluation window
        #                   ┌──┴──┐
        #       ┌─────┬─────┬─────┬─────┐    offset groupers
        #     ┌─┴───┬─┴───┬─┴───┬─┴───┬─┘0 │ by sampling interval
        #   ┌─┴───┬─┴───┬─┴───┬─┴───┬─┘1   │
        #   └─────┴─────┴─────┴─────┘2     ▼
        #         ┌─┬─┬─┬─┬─┬─┬─┬─┬─┬─┬─┐    combine into
        #         └─┴─┴─┴─┴─┴─┴─┴─┴─┴2┴1┘0   final time series
        #
        grouper = pd.Grouper(  # type: ignore  # mypy finds the wrong Grouper
            freq=evaluation_window,
            origin=end_time,
            offset=-offset,
            # Each point in timeseries will be labeled by the end instant of
            # its evaluation window.
            label="right",
            sort=False,
        )
        time_stop = end_time - offset
        if divisible:
            time_start = start_time - evaluation_window
        else:
            time_start = time_stop - evaluation_window
        yield (
            time_start,
            time_stop,
            grouper,
        )


def _results(
    calcs: Iterable[Metric],
    dataframe: pd.DataFrame,
    start_time: datetime,
    end_time: datetime,
    evaluation_window: timedelta,
    sampling_interval: timedelta,
) -> Iterator[pd.DataFrame]:
    """
    Yields metric results for each data point in time series.
    """
    yield pd.DataFrame()
    calculate_metrics = partial(
        multi_calculate,
        calcs=calcs,
    )
    # pandas time indexing is end-inclusive
    result_slice = slice(start_time, end_time)
    for (
        time_start,  # inclusive
        time_stop,  # exclusive
        group,
    ) in _groupers(
        start_time=start_time,
        end_time=end_time,
        evaluation_window=evaluation_window,
        sampling_interval=sampling_interval,
    ):
        row_start, row_stop = row_interval_from_sorted_time_index(
            time_index=dataframe.index,
            time_start=time_start,  # inclusive
            time_stop=time_stop,  # exclusive
        )
        # pandas row indexing is stop-exclusive
        row_slice = slice(row_start, row_stop)
        filtered = dataframe.iloc[row_slice, :]
        res = filtered.groupby(
            group,
            group_keys=True,
        ).apply(
            calculate_metrics,
        )

        # NB: on ubuntu, we lose the timezone information when there is no data
        if res.index.tzinfo is None:  # type: ignore
            res = res.set_axis(
                res.index.tz_localize(  # type: ignore
                    timezone.utc,
                ),
                axis=0,
                copy=False,
            )

        yield res.loc[result_slice, :]



from typing import Optional

import strawberry

from ..interceptor import GqlValueMediator
from .Dimension import Dimension


@strawberry.type
class DimensionWithValue:
    """
    Represents the dimension of the model and the string representation of that
    value for a specific event.
    """

    dimension: Dimension
    value: Optional[str] = strawberry.field(
        description="The string representation of the dimension's value",
        default=GqlValueMediator(),
    )
